import 'package:serverpod_cli/src/util/string_manipulation.dart';
import 'package:source_span/source_span.dart';
import 'package:yaml/yaml.dart';

// ignore: implementation_imports
import 'package:yaml/src/equality.dart';

List<String> convertIndexList(String stringifiedFields) {
  return stringifiedFields.split(',').map((field) => field.trim()).toList();
}

T convertToEnum<T extends Enum>({
  required dynamic value,
  required T enumDefault,
  required List<T> enumValues,
}) {
  if (value is! String) return enumDefault;

  return enumValues.firstWhere(
    (v) => v.name.toLowerCase() == value.toLowerCase(),
    orElse: () => enumDefault,
  );
}

T unsafeConvertToEnum<T extends Enum>({
  required dynamic value,
  required List<T> enumValues,
}) {
  return enumValues.firstWhere(
    (v) => v.name.toLowerCase() == value.toLowerCase(),
  );
}

typedef DeepNestedNodeHandler = YamlMap Function(
    String? content, SourceSpan span);

YamlMap convertStringifiedNestedNodesToYamlMap(
  String? content,
  SourceSpan span, {
  String? firstKey,
  void Function(String key, SourceSpan? span)? onDuplicateKey,
  void Function(String key, SourceSpan? span)? onNegatedKeyWithValue,
}) {
  var stringifiedNodes = _extractStringifiedNodes(content);

  Map<dynamic, YamlNode> initNodes = _extractInitialNode(
    firstKey,
    stringifiedNodes,
    content,
    span,
  );

  var startNodeIndex = initNodes.length;
  var fieldKeyValuePairs = _extractKeyValuePairs(
    stringifiedNodes.skip(startNodeIndex),
    content,
    span,
    onNegatedKeyWithValue: onNegatedKeyWithValue,
    handleDeepNestedNodes: (nestedContent, nestedSpan) {
      // recursion
      return convertStringifiedNestedNodesToYamlMap(
        nestedContent,
        nestedSpan,
        onDuplicateKey: onDuplicateKey,
        onNegatedKeyWithValue: onNegatedKeyWithValue,
      );
    },
  );

  var duplicates = _findDuplicateKeys(fieldKeyValuePairs);
  for (var duplicate in duplicates) {
    onDuplicateKey?.call(
      duplicate,
      _extractSubSpan(content, span, duplicate),
    );
  }

  Map<dynamic, YamlNode> internalNodes = fieldKeyValuePairs.fold(
    initNodes,
    (aggregate, pair) => {...aggregate, ...pair},
  );

  // deepEqualsMap is needed to be able to compare YamlScala as keys
  var nodes = deepEqualsMap<dynamic, YamlNode>();
  nodes.addAll(internalNodes);

  return YamlMap.internal(nodes, span, CollectionStyle.ANY);
}

Map<dynamic, YamlNode> _extractInitialNode(
  String? firstKey,
  List<String> options,
  String? content,
  SourceSpan span,
) {
  if (firstKey == null) return {};

  var initRawValue = options.isNotEmpty ? options.first : null;
  return _createdYamlScalarNode(
    firstKey,
    initRawValue,
    _extractSubSpan(content, span, initRawValue),
  );
}

List<String> _extractStringifiedNodes(String? input) {
  if (input == null) return [];

  // Split on comma, but not if the comma is inside < > or ( )
  return splitIgnoringBrackets(input);
}

Iterable<Map<YamlScalar, YamlNode>> _extractKeyValuePairs(
  Iterable<String> fieldOptions,
  String? content,
  SourceSpan span, {
  void Function(String key, SourceSpan? span)? onNegatedKeyWithValue,
  required DeepNestedNodeHandler handleDeepNestedNodes,
}) {
  if (content == null) return [];

  List<Map<YamlScalar, YamlNode>> fieldPairs = [];

  // This must be a for loop and can never be converted to a map.
  // There seems to be some type of bug (in dart?) that causes the loop
  // to run several times if done with a map. There is also different behaviors
  // between debug and normal runtime.
  for (var stringifiedKeyValuePair in fieldOptions) {
    var keyValueSpan = _extractSubSpan(content, span, stringifiedKeyValuePair);

    if (_hasNestedStringifiedValues(stringifiedKeyValuePair)) {
      var nestedComponents =
          stringifiedKeyValuePair.replaceAll(')', '').split('(');

      var key = nestedComponents.first;
      var stringifiedContent = nestedComponents.last;

      if (stringifiedContent == '') {
        fieldPairs.add(_createdYamlScalarNode(
          key,
          null,
          keyValueSpan,
        ));
        continue;
      } else {
        var nestedSpan = _extractSubSpan(content, span, stringifiedContent);
        var nodeMap = handleDeepNestedNodes(stringifiedContent, nestedSpan);

        fieldPairs.add(_createYamlMapNode(key, nodeMap, keyValueSpan));
        continue;
      }
    }

    List<String> keyValuePair = stringifiedKeyValuePair.split('=');

    String key = keyValuePair.first;
    String? value = keyValuePair.length == 2 ? keyValuePair.last : null;

    if (stringifiedKeyValuePair.contains('=') && key.startsWith('!')) {
      onNegatedKeyWithValue?.call(
        key,
        _extractSubSpan(keyValueSpan.text, keyValueSpan, key),
      );
    }

    if (!stringifiedKeyValuePair.contains('=')) {
      if (key.startsWith('!')) {
        key = key.substring(1);
        value = 'false';
      } else {
        value = 'true';
      }
    }

    fieldPairs.add(_createdYamlScalarNode(
      key,
      value,
      keyValueSpan,
    ));
    continue;
  }

  return fieldPairs;
}

bool _hasNestedStringifiedValues(String stringifiedKeyValuePair) =>
    stringifiedKeyValuePair.contains('(');

SourceSpan _extractSubSpan(
  String? content,
  SourceSpan contentSpan,
  String? subContent,
) {
  if (content == null) return contentSpan;
  if (subContent == null) return contentSpan;

  var start = _findStartIndex(content, subContent);
  if (start == -1) return contentSpan;

  var end = start + subContent.length;
  return contentSpan.subspan(start, end);
}

int _findStartIndex(String fullString, String subString) {
  if (subString == '') return fullString.length;
  return fullString.lastIndexOf(subString);
}

Set<String> _findDuplicateKeys(Iterable<Map<YamlScalar, dynamic>> list) {
  var seenStrings = <String>{};
  var duplicates = <String>{};

  for (var pair in list) {
    var v = pair.entries.first;
    if (seenStrings.contains(v.key.value)) {
      duplicates.add(v.key.value);
    } else {
      seenStrings.add(v.key.value);
    }
  }

  return duplicates;
}

Map<YamlScalar, YamlScalar> _createdYamlScalarNode(
  String rawKey,
  String? rawValue,
  SourceSpan span,
) {
  var trimmedKey = rawKey.trim();
  var keySpan = _extractSubSpan(span.text, span, trimmedKey);
  var key = YamlScalar.internalWithSpan(trimmedKey, keySpan);

  var trimmedValue = rawValue?.trim();
  var valueSpan = _extractSubSpan(span.text, span, trimmedValue);
  var value = YamlScalar.internalWithSpan(trimmedValue, valueSpan);

  return {key: value};
}

Map<YamlScalar, YamlMap> _createYamlMapNode(
  String rawKey,
  YamlMap value,
  SourceSpan span,
) {
  var trimmedKey = rawKey.trim();
  var keySpan = _extractSubSpan(span.text, span, trimmedKey);
  var key = YamlScalar.internalWithSpan(trimmedKey, keySpan);

  return {key: value};
}
